/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2006,2007, ... ,2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/random.h>
#include <linux/math64.h>
#include "cpmac_if.h"
#include "cpmac_debug.h"
#include "cpphy_const.h"
#include "cpphy_types.h"
#include "cpmac_test.h"


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
extern unsigned char get_network_dev_num(struct net_device *dev);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#define AVM_CPMAC_TEST_BUFFER_SIZE 64000

static char ethernet_header[] = { '\xff', '\xff', '\xff', '\xff', '\xff', '\xff',
                                  '\x00', '\x01', '\x02', '\x03', '\x04', '\x05',
                                  '\xfe', '\xdc'
                                };
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static struct sk_buff *cpmac_test_generate_packet(unsigned char port,
                                                  unsigned int size,
                                                  unsigned int counter) {
    struct sk_buff *skb;
    unsigned int left = AVM_CPMAC_TEST_BUFFER_SIZE - 1 - cpmac_global.lan_test.buffer_offset;

    if(size < 60) {
        DEB_WARN("[%s] Size %u too small for an ethernet packet\n", __FUNCTION__, size);
        return NULL;
    }
    if((skb = alloc_skb(size, GFP_KERNEL)) == NULL) {
        DEB_WARN("[%s] Could not allocate memory for skb\n", __FUNCTION__);
        return NULL;
    }
    skb->uniq_id = 0x47110815;
    skb_put(skb, size);
    memcpy(skb->data, ethernet_header, 14);
    skb->data[11] += port;
    *((unsigned int *) (skb->data + 14)) = counter;

    size -= 32;
    if(size < left) {
        memcpy(skb->data + 32, cpmac_global.lan_test.buffer + cpmac_global.lan_test.buffer_offset, size);
        cpmac_global.lan_test.buffer_offset += size;
    } else {
        memcpy(skb->data + 32, cpmac_global.lan_test.buffer + cpmac_global.lan_test.buffer_offset, left);
        memcpy(skb->data + 32 + left, cpmac_global.lan_test.buffer, size - left);
        cpmac_global.lan_test.buffer_offset = size - left;
    }

    return skb;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_test_interprete_packet(struct sk_buff *skb, unsigned char instance) {
    unsigned char testport = cpmac_global.lan_test.map_instance_to_testport[instance];
    int counter_diff;
    unsigned int counter;

    if(   (skb->data[12] != 0xfe)
       || (skb->data[13] != 0xdc)) {
        kfree_skb(skb);
        return;
    }

    counter = *((unsigned int *) (skb->data + 14));
    counter_diff = counter - cpmac_global.lan_test.rx_expected[testport];
    if(counter_diff) {
        if(counter_diff > 0) {
            cpmac_global.lan_test.lost_counter[testport] += counter_diff;
        } else {
            DEB_ERR("[%s] Testport %u: Received duplicate packet %u. Expected %u!\n", __FUNCTION__, 
                    testport, counter, cpmac_global.lan_test.rx_expected[testport]);
        }
    }
    cpmac_global.lan_test.rx_expected[testport] = counter + 1;
    kfree_skb(skb);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_lan_test_work(struct work_struct *workstruct) {
    cpmac_lan_test_work_t *work;
    unsigned char testport;
    unsigned int i;
    struct sk_buff *skb;
    unsigned long delay = 0;
    
    assert(workstruct);
    work = container_of((struct delayed_work *) workstruct, 
                        cpmac_lan_test_work_t,
                        work);
    assert(work);
    testport = work->testport;
    for(i = 0; i < 10; i++) {
        if(cpmac_global.lan_test.tx_counter[testport] >= cpmac_global.lan_test.no_of_testpkts) {
            DEB_TEST("[%s] Sending for testport %u done.\n", __FUNCTION__, testport);
            if(cpmac_global.lan_test.tx_counter[testport ? 0 : 1] >= cpmac_global.lan_test.no_of_testpkts) {
                cpmac_global.lan_test.test_end = jiffies;
            }
            return;
        }

        skb = cpmac_test_generate_packet(testport,
                                         cpmac_global.lan_test.pkt_size,
                                         cpmac_global.lan_test.tx_counter[testport]);
        if(skb) {
            skb->dev = cpmac_global.instance[cpmac_global.lan_test.ports[testport]].dev;
            if(0 != dev_queue_xmit(skb)) {;
                delay = HZ / 20;
                break;
            }
            cpmac_global.lan_test.tx_counter[testport]++;
        } else {
            delay = HZ / 20;
            break;
        }
    }
    queue_delayed_work(cpmac_global.lan_test.workqueue, &cpmac_global.lan_test.work[testport].work, delay);
    schedule();
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_test_start(unsigned int pkt_size, unsigned int pkt_count) {
    unsigned char i;

    if((pkt_size > 1480) || (pkt_size < 64)) {
        DEB_ERR("[%s] Wrong packet size %u requested. Must be between 64 and 1480.\n",
                __FUNCTION__, pkt_size);
        return CPMAC_ERR_EXCEEDS_LIMIT;
    }
    cpmac_global.lan_test.pkt_size       = pkt_size - 4; /* Checksum will be there anyway */
    cpmac_global.lan_test.no_of_testpkts = pkt_count;
    cpmac_global.lan_test.test_start = jiffies;
    cpmac_global.lan_test.test_end = 0;
    for(i = 0; i < 2; i++) {
        cpmac_global.lan_test.tx_counter[i] = 0;
        cpmac_global.lan_test.rx_expected[i] = 0;
        cpmac_global.lan_test.lost_counter[i] = 0;
        cpmac_global.lan_test.work[i].testport = i;
        INIT_DELAYED_WORK(&cpmac_global.lan_test.work[i].work, cpmac_lan_test_work);
        queue_delayed_work(cpmac_global.lan_test.workqueue, &cpmac_global.lan_test.work[i].work, 0);
    }
    DEB_SUPPORT("LAN test started\n");

    return CPMAC_ERR_NOERR;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int cpmac_test_lan_result(char *buffer, int count) {
    int bufptr = 0;
    unsigned char i;

    /* Wait, until the sending is done */
    while(   (cpmac_global.lan_test.tx_counter[0] != cpmac_global.lan_test.no_of_testpkts)
          || (cpmac_global.lan_test.tx_counter[1] != cpmac_global.lan_test.no_of_testpkts)) {
        ssleep(1);
    }
    DEB_SUPPORT("Sending done, waiting 3s to ensure completion of rx activities\n");
    ssleep(3);

    bufptr += snprintf(buffer + bufptr, count - bufptr, "LAN test statistics:\n");
    for(i = 0; i < 2; i++) {
        cpmac_global.lan_test.lost_counter[i] += cpmac_global.lan_test.no_of_testpkts - cpmac_global.lan_test.rx_expected[i];
        bufptr += snprintf(buffer + bufptr, count - bufptr, 
                           "  Packets on port %u: sent %u  lost %u\n",
                           cpmac_global.lan_test.ports[i],
                           cpmac_global.lan_test.tx_counter[i],
                           cpmac_global.lan_test.lost_counter[i]);
    }
    bufptr += snprintf(buffer + bufptr, count - bufptr, 
                       "  Packets total: sent %u  lost %u  => %lu / 1.000.000 packet loss\n",
                       2 * cpmac_global.lan_test.no_of_testpkts,
                       cpmac_global.lan_test.lost_counter[0] + cpmac_global.lan_test.lost_counter[1],
                         ((1000000 / 2) * (unsigned long) (cpmac_global.lan_test.lost_counter[0] + cpmac_global.lan_test.lost_counter[1])) 
                       / (unsigned long) cpmac_global.lan_test.no_of_testpkts);
    /*--- { ---*/
        /*--- unsigned long time_in_ms = ((cpmac_global.lan_test.test_end - cpmac_global.lan_test.test_start) * 1000lu) / HZ; ---*/
        /*--- DEB_SUPPORT("  Test ran %lu ms, throughput %llu kbps\n", ---*/
                    /*--- time_in_ms, div64_u64(8lu * (unsigned long long) (cpmac_global.lan_test.no_of_testpkts / 1000) * (unsigned long) cpmac_global.lan_test.pkt_size, time_in_ms) / 1000); ---*/
    /*--- } ---*/
    return bufptr;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_test_init(unsigned char instance, unsigned char instance2) {
    unsigned int i;

    if(   (instance >= AVM_CPMAC_MAX_INSTANCES) 
       || (instance2 >= AVM_CPMAC_MAX_INSTANCES)
       || (instance == instance2)) {
        DEB_ERR("[%s] Illegal instances (%u, %u) given\n", __FUNCTION__, instance, instance2);
        return CPMAC_ERR_ILL_HANDLE;
    }
    if(cpmac_global.lan_test.buffer) {
        DEB_ERR("[%s] Test buffer is already allocated!\n", __FUNCTION__);
        return CPMAC_ERR_ALREADY_REGISTERED;
    }

    cpmac_global.lan_test.workqueue = create_singlethread_workqueue("CPMAC lan test");

    DEB_INFO("[%s] Init test for devices '%s'(%u) and '%s'(%u)\n", __FUNCTION__, 
             cpmac_global.instance[instance].dev->name, instance,
             cpmac_global.instance[instance2].dev->name, instance2);

    cpmac_global.lan_test.buffer = kmalloc(AVM_CPMAC_TEST_BUFFER_SIZE, GFP_KERNEL);
    if(!cpmac_global.lan_test.buffer) {
        DEB_ERR("[%s] Out of memory for test buffer!\n", __FUNCTION__);
        return CPMAC_ERR_NOMEM;
    }
    for(i = 0; i < AVM_CPMAC_TEST_BUFFER_SIZE / 4; i++) {
        ((unsigned int *) cpmac_global.lan_test.buffer)[i] = random32();
    }

    for(i = 0; i < AVM_CPMAC_MAX_INSTANCES; i++) {
        cpmac_global.lan_test.map_instance_to_testport[i] = 0;
    }
    cpmac_global.lan_test.map_instance_to_testport[instance] = 0;
    cpmac_global.lan_test.map_instance_to_testport[instance2] = 1;
    cpmac_global.instance[instance].is_in_testmode  = 1;
    cpmac_global.instance[instance2].is_in_testmode = 1;
    cpmac_global.lan_test.ports[0] = instance;
    cpmac_global.lan_test.ports[1] = instance2;

    return CPMAC_ERR_NOERR;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_test_deinit(void) {
    unsigned char instance;

    if(!cpmac_global.lan_test.buffer) {
        DEB_ERR("[%s] Test buffer is not allocated!\n", __FUNCTION__);
        return CPMAC_ERR_CHAN_NOT_OPEN;
    }
    for(instance = 0; instance < AVM_CPMAC_MAX_INSTANCES; instance++) {
        cpmac_global.instance[instance].is_in_testmode  = 0;
    }
    kfree(cpmac_global.lan_test.buffer);
    cpmac_global.lan_test.buffer = NULL;

    flush_workqueue(cpmac_global.lan_test.workqueue);
    destroy_workqueue(cpmac_global.lan_test.workqueue);

    return CPMAC_ERR_NOERR;
}

